==> imspector-0.5.tar.gz: imspector-0.5/yahooprotocolplugin.cpp <==
/* IMSpector - Instant Messenger Transparent Proxy Service
 * http://www.imspector.org/
 * (c) Lawrence Manning <lawrence@aslak.net>, 2006
 *          
 * Released under the GPL v2. */

#include "imspector.h"

#define PLUGIN_NAME "Yahoo IMSpector protocol plugin"
#define PROTOCOL_NAME "Yahoo"
#define PROTOCOL_PORT 5050

extern "C"
{
	bool initprotocolplugin(struct protocolplugininfo &pprotocolplugininfo,
		class Options &options, bool debugmode);
	void closeprotocolplugin(void);
	int processpacket(bool outgoing, class Socket &incomingsock, char *replybuffer,
		int *replybufferlength, std::vector<struct imevent> &imevents, std::string &clientaddress);
};

#pragma pack(2)

struct header
{
	char ymsg[4];
	uint32_t version;
	uint16_t pkt_len;
	uint16_t service;
	uint32_t status;
	uint32_t session_id;
};

#pragma pack()

struct tagvalue
{
	std::string text;
	struct messageextent messageextent;
};

#define YAHOO_SERVICE_MESSAGE 0x06
#define YAHOO_SERVICE_GROUP_MESSAGE 0x1d
#define YAHOO_SERVICE_FILETRANSFER 0x46
#define YAHOO_SERVICE_P2PFILEXFER 0x4d
#define YAHOO_SERVICE_MISC 0x4b

int packetcount = 0;
bool tracing = false;
bool localdebugmode = false;

int gettagsandvalues(uint8_t *buffer, int length, std::map<std::string, struct tagvalue> &tagvalues);

bool initprotocolplugin(struct protocolplugininfo &protocolplugininfo,
	class Options &options, bool debugmode)
{
	if (options["yahoo_protocol"] != "on") return false;

	localdebugmode = debugmode;
	
	protocolplugininfo.pluginname = PLUGIN_NAME;
	protocolplugininfo.protocolname = PROTOCOL_NAME;
	protocolplugininfo.port = htons(PROTOCOL_PORT);
	
	if (options["yahoo_trace"] == "on") tracing = true;
	
	return true;
}

void closeprotocolplugin(void)
{
	return;
}

/* The main plugin function. See protocolplugin.cpp. */
int processpacket(bool outgoing, class Socket &incomingsock, char *replybuffer, 
	int *replybufferlength, std::vector<struct imevent> &imevents, std::string &clientaddress)
{
	struct header header;
	memset(&header, 0, sizeof(struct header));

	/* Get the fix sized header and ensure endianess. */	
	if (!incomingsock.recvalldata((char *) &header, sizeof(struct header))) return 1;

	memcpy(replybuffer, &header, sizeof(struct header));
	*replybufferlength = sizeof(struct header);

	header.version = ntohl(header.version);
	header.pkt_len = ntohs(header.pkt_len);
	header.service = ntohs(header.service);
	header.status = ntohs(header.status);
	header.session_id = ntohl(header.session_id);
	
	debugprint(localdebugmode, PROTOCOL_NAME ": ID: %c %c %c %c verion: %d service: %04x pkt_len: %d",
		header.ymsg[0], header.ymsg[1],	header.ymsg[2], header.ymsg[3], 
		header.version, header.service, header.pkt_len);
		
	uint8_t buffer[BUFFER_SIZE];
	
	/* Get the following data assuming the header indicates there is some. */
	if (header.pkt_len)
	{
		if (!incomingsock.recvalldata((char *) buffer, header.pkt_len)) return 1;
		
		memcpy(replybuffer + sizeof(struct header), buffer, header.pkt_len);
		*replybufferlength += header.pkt_len;
	}

	/* Split out the buffer into a list of tag and values. */
	std::map<std::string, struct tagvalue> tagvalues;
	gettagsandvalues(buffer, header.pkt_len, tagvalues);

	/* Simple stuff: build the imevent and push it onto the list. */
	struct imevent imevent;
	
	imevent.outgoing = outgoing;
	imevent.type = TYPE_NULL;
	imevent.filtered = false;
	imevent.messageextent.start = 0;
	imevent.messageextent.length = 0;
	
	if (header.service == YAHOO_SERVICE_MESSAGE)
	{
		imevent.type = TYPE_MSG;
		imevent.eventdata = tagvalues["14"].text;
		
		imevent.messageextent = tagvalues["14"].messageextent;
	}
	else if (header.service == YAHOO_SERVICE_GROUP_MESSAGE)
	{
		imevent.type = TYPE_MSG;
		if (outgoing)
			imevent.eventdata = tagvalues["14"].text;
		else
			imevent.eventdata = tagvalues["3"].text + ": " + tagvalues["14"].text;
		
		imevent.messageextent = tagvalues["14"].messageextent;
	}	
	else if (header.service == YAHOO_SERVICE_P2PFILEXFER)
	{
		imevent.type = TYPE_FILE;
		imevent.eventdata = tagvalues["27"].text + " " + tagvalues["28"].text + " bytes";
	}
	else if (header.service == YAHOO_SERVICE_MISC)
	{
		if (tagvalues["49"].text == "TYPING")
		{
			imevent.type = TYPE_TYPING;
			imevent.eventdata = "";
		}
		if (tagvalues["49"].text == "WEBCAMINVITE")
		{
			imevent.type = TYPE_WEBCAM;
			imevent.eventdata = "";
		}
	}
	
	if (imevent.type != TYPE_NULL)
	{
		imevent.timestamp = time(NULL);
		imevent.clientaddress = clientaddress;
		imevent.protocolname = PROTOCOL_NAME;

		if (header.service == YAHOO_SERVICE_GROUP_MESSAGE)
		{
			if (outgoing)
			{
				imevent.localid = tagvalues["1"].text;
				imevent.remoteid = tagvalues["57"].text;
			}
			else
			{
				imevent.localid = tagvalues["1"].text;
				imevent.remoteid = tagvalues["57"].text;
			}
		}
		else	
		{
			if (outgoing)
			{
				imevent.localid = tagvalues["1"].text;
				imevent.remoteid = tagvalues["5"].text;
			}
			else
			{
				imevent.localid = tagvalues["5"].text;
				imevent.remoteid = tagvalues["4"].text;
			}
		}

		std::transform(imevent.localid.begin(), imevent.localid.end(), imevent.localid.begin(), tolower);
		std::transform(imevent.remoteid.begin(), imevent.remoteid.end(), imevent.remoteid.begin(), tolower);
		
		imevents.push_back(imevent);
	}
	
	/* Write out trace packets if enabled. TODO: This should probably be generic
	 * code somewhere up the call-chain. */
	if (tracing)
	{
		char filename[STRING_SIZE];
		memset(filename, 0, STRING_SIZE);
		
		snprintf(filename, STRING_SIZE - 1, TRACE_DIR "/yahoo.%d.%d", getpid(), packetcount);

		int fd = -1;
		if ((fd = creat(filename, 0600)) > 0)
		{
			write(fd, replybuffer, *replybufferlength);
			close(fd);
		}
	}
	
	packetcount++;
	
	return 0;
}

/* Builds up a map of tag and values using the c080 markers. */
int gettagsandvalues(uint8_t *buffer, int length, std::map<std::string, struct tagvalue> &tagvalues)
{
	int counter = 0;
	uint8_t *p = (uint8_t *) buffer;
	
	while (p - buffer < length)
	{
		std::string tag; struct tagvalue tagvalue;	
		
		while (!(p[0] == 0xc0 && p[1] == 0x80))
		{
			if (p - buffer >= length) break;
			tag += *p;
			p++;
		}
		p += 2;
		tagvalue.messageextent.start = p - buffer + sizeof(struct header);
		tagvalue.messageextent.length = 0;
		while (!(p[0] == 0xc0 && p[1] == 0x80))
		{
			if (p - buffer >= length) break;
			tagvalue.text += *p;
			tagvalue.messageextent.length++;
			p++;
		}
		p += 2;
		
		tagvalues[tag] = tagvalue;
		counter++;
		
		debugprint(localdebugmode, PROTOCOL_NAME ": Tag: %s Value: %s", tag.c_str(), tagvalue.text.c_str());
	}
	
	return counter;
}

==> imspector-0.7.tar.gz: imspector-0.7/yahooprotocolplugin.cpp <==
/* IMSpector - Instant Messenger Transparent Proxy Service
 * http://www.imspector.org/
 * (c) Lawrence Manning <lawrence@aslak.net>, 2006
 *          
 * Released under the GPL v2. */

#include "imspector.h"

#define PLUGIN_NAME "Yahoo IMSpector protocol plugin"
#define PROTOCOL_NAME "Yahoo"
#define PROTOCOL_PORT 5050

extern "C"
{
	bool initprotocolplugin(struct protocolplugininfo &pprotocolplugininfo,
		class Options &options, bool debugmode);
	void closeprotocolplugin(void);
	int processpacket(bool outgoing, class Socket &incomingsock, char *replybuffer,
		int *replybufferlength, std::vector<struct imevent> &imevents, std::string &clientaddress);
};

#pragma pack(2)

struct header
{
	char ymsg[4];
	uint32_t version;
	uint16_t pkt_len;
	uint16_t service;
	uint32_t status;
	uint32_t session_id;
};

#pragma pack()

struct tagvalue
{
	std::string text;
	struct messageextent messageextent;
};

#define YAHOO_SERVICE_MESSAGE 0x06
#define YAHOO_SERVICE_GROUP_MESSAGE 0x1d
#define YAHOO_SERVICE_FILETRANSFER 0x46
#define YAHOO_SERVICE_P2PFILEXFER 0x4d
#define YAHOO_SERVICE_MISC 0x4b

int packetcount = 0;
bool tracing = false;
bool localdebugmode = false;

int gettagsandvalues(uint8_t *buffer, int length, std::map<std::string, struct tagvalue> &tagvalues);

bool initprotocolplugin(struct protocolplugininfo &protocolplugininfo,
	class Options &options, bool debugmode)
{
	if (options["yahoo_protocol"] != "on") return false;

	localdebugmode = debugmode;
	
	protocolplugininfo.pluginname = PLUGIN_NAME;
	protocolplugininfo.protocolname = PROTOCOL_NAME;
	protocolplugininfo.port = htons(PROTOCOL_PORT);
	
	if (options["yahoo_trace"] == "on") tracing = true;
	
	return true;
}

void closeprotocolplugin(void)
{
	return;
}

/* The main plugin function. See protocolplugin.cpp. */
int processpacket(bool outgoing, class Socket &incomingsock, char *replybuffer, 
	int *replybufferlength, std::vector<struct imevent> &imevents, std::string &clientaddress)
{
	struct header header;
	memset(&header, 0, sizeof(struct header));

	/* Get the fix sized header and ensure endianess. */	
	if (!incomingsock.recvalldata((char *) &header, sizeof(struct header))) return 1;

	memcpy(replybuffer, &header, sizeof(struct header));
	*replybufferlength = sizeof(struct header);

	header.version = ntohl(header.version);
	header.pkt_len = ntohs(header.pkt_len);
	header.service = ntohs(header.service);
	header.status = ntohs(header.status);
	header.session_id = ntohl(header.session_id);
	
	debugprint(localdebugmode, PROTOCOL_NAME ": ID: %c %c %c %c verion: %d service: %04x pkt_len: %d",
		header.ymsg[0], header.ymsg[1],	header.ymsg[2], header.ymsg[3], 
		header.version, header.service, header.pkt_len);
		
	uint8_t buffer[BUFFER_SIZE];
	
	/* Get the following data assuming the header indicates there is some. */
	if (header.pkt_len)
	{
		if (!incomingsock.recvalldata((char *) buffer, header.pkt_len)) return 1;
		
		memcpy(replybuffer + sizeof(struct header), buffer, header.pkt_len);
		*replybufferlength += header.pkt_len;
	}

	/* Split out the buffer into a list of tag and values. */
	std::map<std::string, struct tagvalue> tagvalues;
	gettagsandvalues(buffer, header.pkt_len, tagvalues);

	/* Simple stuff: build the imevent and push it onto the list. */
	struct imevent imevent;
	
	imevent.outgoing = outgoing;
	imevent.type = TYPE_NULL;
	imevent.filtered = false;
	imevent.messageextent.start = 0;
	imevent.messageextent.length = 0;
	
	if (header.service == YAHOO_SERVICE_MESSAGE)
	{
		imevent.type = TYPE_MSG;
		imevent.eventdata = tagvalues["14"].text;
		
		imevent.messageextent = tagvalues["14"].messageextent;
	}
	else if (header.service == YAHOO_SERVICE_GROUP_MESSAGE)
	{
		imevent.type = TYPE_MSG;
		if (outgoing)
			imevent.eventdata = tagvalues["14"].text;
		else
			imevent.eventdata = tagvalues["3"].text + ": " + tagvalues["14"].text;
		
		imevent.messageextent = tagvalues["14"].messageextent;
	}	
	else if (header.service == YAHOO_SERVICE_P2PFILEXFER)
	{
		imevent.type = TYPE_FILE;
		imevent.eventdata = tagvalues["27"].text + " " + tagvalues["28"].text + " bytes";
	}
	else if (header.service == YAHOO_SERVICE_MISC)
	{
		if (tagvalues["49"].text == "TYPING")
		{
			imevent.type = TYPE_TYPING;
			imevent.eventdata = "";
		}
		if (tagvalues["49"].text == "WEBCAMINVITE")
		{
			imevent.type = TYPE_WEBCAM;
			imevent.eventdata = "";
		}
	}
	
	if (imevent.type != TYPE_NULL)
	{
		imevent.timestamp = time(NULL);
		imevent.clientaddress = clientaddress;
		imevent.protocolname = PROTOCOL_NAME;

		if (header.service == YAHOO_SERVICE_GROUP_MESSAGE)
		{
			if (outgoing)
			{
				imevent.localid = tagvalues["1"].text;
				imevent.remoteid = tagvalues["57"].text;
			}
			else
			{
				imevent.localid = tagvalues["1"].text;
				imevent.remoteid = tagvalues["57"].text;
			}
		}
		else	
		{
			if (outgoing)
			{
				imevent.localid = tagvalues["1"].text;
				imevent.remoteid = tagvalues["5"].text;
			}
			else
			{
				imevent.localid = tagvalues["5"].text;
				imevent.remoteid = tagvalues["4"].text;
			}
		}

		std::transform(imevent.localid.begin(), imevent.localid.end(), imevent.localid.begin(), tolower);
		std::transform(imevent.remoteid.begin(), imevent.remoteid.end(), imevent.remoteid.begin(), tolower);
		
		imevents.push_back(imevent);
	}
	
	/* Write out trace packets if enabled. */
	if (tracing) tracepacket("yahoo", packetcount, replybuffer, *replybufferlength);
	
	packetcount++;
	
	return 0;
}

/* Builds up a map of tag and values using the c080 markers. */
int gettagsandvalues(uint8_t *buffer, int length, std::map<std::string, struct tagvalue> &tagvalues)
{
	int counter = 0;
	uint8_t *p = (uint8_t *) buffer;
	
	while (p - buffer < length)
	{
		std::string tag; struct tagvalue tagvalue;	
		
		while (!(p[0] == 0xc0 && p[1] == 0x80))
		{
			if (p - buffer >= length) break;
			tag += *p;
			p++;
		}
		p += 2;
		tagvalue.messageextent.start = p - buffer + sizeof(struct header);
		tagvalue.messageextent.length = 0;
		while (!(p[0] == 0xc0 && p[1] == 0x80))
		{
			if (p - buffer >= length) break;
			tagvalue.text += *p;
			tagvalue.messageextent.length++;
			p++;
		}
		p += 2;
		
		tagvalues[tag] = tagvalue;
		counter++;
		
		debugprint(localdebugmode, PROTOCOL_NAME ": Tag: %s Value: %s", tag.c_str(), tagvalue.text.c_str());
	}
	
	return counter;
}
